/*****************************************************************************
 * dshowbridge.cpp : DirectShow access module for vlc
 *****************************************************************************
 * Copyright (C) 2007 Sensoray Company Inc.
 * Author: Sensoray Company Inc.
 * Based on dshow.cpp with Copyright below. 
 * (completely re-written for standalone VLC Direct MUX + FileWriting filter
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version. 
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston MA 02110-1301, USA.
 *****************************************************************************/
/*****************************************************************************
 * Preamble
 *****************************************************************************/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <vlc/vlc.h>
#include <vlc/input.h>
#include <vlc/vout.h>

#include "common.h"
#include "filter.h"
#include <dvdmedia.h>


#define STREAM_MASK_VIDEO 2
#define STREAM_MASK_AUDIO 1

/*****************************************************************************
 * Access: local prototypes
 *****************************************************************************/
static int Demux       ( demux_t * );
static int DemuxControl( demux_t *, int, va_list );
static int BuildGraph( vlc_object_t *p_this, access_sys_t *p_sys);
static void ParseMRL( vlc_object_t *p_this );

/*****************************************************************************
 * Module descriptor
 *****************************************************************************/
#define CACHING_TEXT N_("Caching value in ms")
#define CACHING_LONGTEXT N_( \
    "Caching value for DirectShow streams. " \
    "This value should be set in milliseconds." )
#define STREAMS_TEXT N_("stream mask")
#define STREAMS_LONGTEXT N_( \
    "Set the stream mask(1 audio, 2 video) " )
#define SYSTEM_IN_TEXT N_("System")
#define SYSTEM_IN_LONGTEXT N_( "NTSC 2 default, PAL-1")

static int  CommonOpen ( vlc_object_t *, access_sys_t *, vlc_bool_t );
static void CommonClose( vlc_object_t *, access_sys_t * );
static int  DemuxOpen  ( vlc_object_t * );
static void DemuxClose ( vlc_object_t * );


vlc_module_begin();
    set_shortname( _("dshowbridge") );
    set_description( _("Sensoray Directshow bridge module") );
    set_category( CAT_INPUT );
    set_subcategory( SUBCAT_INPUT_ACCESS );

    add_integer( "dshowbridge-caching", ( (mtime_t)(.5*CLOCK_FREQ)/ 1000) , NULL,
                 CACHING_TEXT, CACHING_LONGTEXT, VLC_TRUE );

    add_integer( "dshowbridge-streams", 3, NULL, STREAMS_TEXT,
                 STREAMS_LONGTEXT, VLC_TRUE );

    add_integer( "dshowbridge-system",  2, NULL, SYSTEM_IN_TEXT,
                 SYSTEM_IN_LONGTEXT, VLC_TRUE );

    set_description( _("Sensoray DirectShow Bridge") );
    add_shortcut( "dshowbridge" );
    set_capability( "access_demux", 0);
    set_callbacks( DemuxOpen, DemuxClose );
vlc_module_end(); 


/*****************************************************************************
 * DemuxOpen: open direct show device as an access_demux module
 *****************************************************************************/
static int DemuxOpen( vlc_object_t *p_this )
{
    demux_t      *p_demux = (demux_t *)p_this;
    access_sys_t *p_sys;
    msg_Dbg( p_this, "Demux Opn...");

    p_sys = (access_sys_t *)malloc( sizeof( access_sys_t ) );
    memset( p_sys, 0, sizeof( access_sys_t ) );
    p_demux->p_sys = (demux_sys_t *)p_sys;
    p_sys->p_graph = (IGraphBuilder *) p_demux->p_vlc->pgraph;
    // we get a pointer to the graph.  so we must addref it. 
    // in demuxclose, we release it.
    // Otherwise there is a race condition
    p_sys->p_graph->AddRef();
    msg_Dbg( p_this, "Demux Open 1 %p...\n", p_demux->p_vlc->pgraph);

    p_sys->b_pts[0] = false;
    p_sys->b_pts[1] = false;
    p_sys->m_vid_pts_times_ten = 0;
    p_sys->m_aud_pts = 0;
    p_sys->m_pinAud = (IPin *) p_demux->p_vlc->pinAud;
    p_sys->m_pinVid = (IPin *) p_demux->p_vlc->pinVid;
    msg_Dbg( p_this, "Demux Open %p %p...", p_sys->m_pinVid, p_sys->m_pinAud);
    if( CommonOpen( p_this, p_sys, VLC_TRUE ) != VLC_SUCCESS )
    {
        msg_Err(p_this, "Sensoray DirectShow bridge open failed");
        CommonClose( p_this, p_sys );
        return VLC_EGENERIC;
    }
    // 
    switch( p_sys->i_compression ) {
        case MPEG4:
        case MPEG4DIVX:
            p_sys->i_fourcc = VLC_FOURCC( 'm', 'p', '4', 'v' ) ;
            break;
        case MOTIONJPEG:
            p_sys->i_fourcc = VLC_FOURCC( 'm', 'j', 'p', 'g' ) ;
            break;            
        case MPEG1:
        case MPEG2:
        default:
            p_sys->i_fourcc = VLC_FOURCC( 'm', 'p', 'g', 'v' ) ;
            break;
    }

    // Declare elementary streams 
    if(p_sys->i_streams & STREAM_MASK_VIDEO) {
        es_format_t fmt;
        es_format_Init( &fmt, VIDEO_ES, p_sys->i_fourcc );
        fmt.video.i_frame_rate = p_sys->i_framerate;
        fmt.video.i_frame_rate_base = 1001;
        fmt.i_bitrate = p_sys->i_vrate;
        msg_Dbg( p_demux, "===added new video es[%4.4s] fr:%d bitrate: %d=======", (char*)&fmt.i_codec, fmt.video.i_frame_rate, fmt.i_bitrate);
        p_sys->p_es_video = es_out_Add( p_demux->out, &fmt );
        if(p_sys->p_es_video == NULL) {
            msg_Err( p_demux, "Sensoray DirectShow Bridge: failed to add VideoES\n");
            return VLC_EGENERIC;
        }
    }

    if(p_sys->i_streams & STREAM_MASK_AUDIO) 
    { 
        // add the raw audio stream
        es_format_t fmt;
        es_format_Init( &fmt, AUDIO_ES, p_sys->i_aud_fourcc);
        fmt.audio.i_channels = p_sys->i_aud_channels;
        fmt.audio.i_rate = p_sys->i_aud_sr;
        fmt.audio.i_bitspersample = p_sys->i_aud_bits;  
        fmt.audio.i_blockalign = fmt.audio.i_channels *
            fmt.audio.i_bitspersample / 8;
        p_sys->i_aud_blockalign = fmt.audio.i_blockalign;
        fmt.i_bitrate = fmt.audio.i_channels * fmt.audio.i_rate *
            fmt.audio.i_bitspersample;
        msg_Dbg( p_demux, "added new audio es %d channels %dHz, streams %d",
                 fmt.audio.i_channels, fmt.audio.i_rate, p_sys->i_streams );
        // add audio to demux
        p_sys->p_es_audio = es_out_Add( p_demux->out, &fmt );
        if(p_sys->p_es_audio == NULL) {
            msg_Err( p_demux, "Sensoray DirectShow Bridge: failed to add audio\n");
            return VLC_EGENERIC;
        }
    }

    es_out_Control( p_demux->out, ES_OUT_SET_MODE, ES_OUT_MODE_AUTO);

    msg_Dbg( p_demux, "dshowbridge success" );
    p_demux->pf_demux   = Demux;
    p_demux->pf_control = DemuxControl;
    p_demux->info.i_update = 0;
    p_demux->info.i_title = 0;
    p_demux->info.i_seekpoint = 0;
    msg_Info( p_this, "Open");
    p_sys->open = 1;
    p_demux->p_vlc->sraybridge_state = 1;
    // let the main module know we are connected.  This allows the directshow 
    return VLC_SUCCESS;
}



/*****************************************************************************
 * ParseMRL: parse the options contained in the MRL
 *****************************************************************************/
static void ParseMRL( vlc_object_t *p_this )
{
    demux_t      *p_demux = (demux_t *) p_this;
    access_sys_t *p_sys = (access_sys_t *)p_demux->p_sys;
    char *psz_dup = strdup( p_demux->psz_path );
    char *psz_parser = psz_dup;
    while( *psz_parser && *psz_parser != ':' )
    {
        psz_parser++;
    }
    if( *psz_parser == ':' )
    {
        /* read options */
        for( ;; )
        {
            *psz_parser++ = '\0';
            if( !strncmp( psz_parser, "system=", strlen( "system=" ) ) )
            {
                p_sys->i_norm = strtol( psz_parser + strlen( "system=" ),
                    &psz_parser, 0 );
                msg_Dbg(p_demux, "system %d\n",p_sys->i_norm);
            }
            else if( !strncmp( psz_parser, "streams=", strlen( "streams=" ) ) )
            {

                p_sys->i_streams = strtol( psz_parser + strlen( "streams=" ),
                    &psz_parser, 0 );
            }
            else
            {
                msg_Warn( p_demux, "unknown option" );
            }

            while( *psz_parser && *psz_parser != ':' )
            {
                psz_parser++;
            }

            if( *psz_parser == '\0' )
            {
                break;
            }
        }
    }
    if( psz_dup ) free( psz_dup );
}

/*****************************************************************************
 * CommonOpen: open direct show device
 *****************************************************************************/
static int CommonOpen( vlc_object_t *p_this, access_sys_t *p_sys,
                       vlc_bool_t b_access_demux )
{
    var_Create( p_this, "dshowbridge-streams", VLC_VAR_INTEGER | VLC_VAR_DOINHERIT );
    var_Create( p_this, "dshowbridge-system", VLC_VAR_INTEGER | VLC_VAR_DOINHERIT );
    msg_Dbg( p_this, "CommonOpen1...");
    p_sys->i_streams = var_GetInteger( p_this, "dshowbridge-streams" );
    // (not currently used)
    p_sys->i_norm =  var_GetInteger( p_this, "dshowbridge-system" );
    // these values are changed based on the directshow connection
    p_sys->i_framerate =  30000; 
    p_sys->i_vrate = 8000000;
    p_sys->i_compression = MPEG1;
    p_sys->i_aud_sr = 48000;
    p_sys->i_aud_bits = 16;
    p_sys->i_aud_channels = 2;
    p_sys->i_aud_fourcc = FOURCC_AUDIO_PCM;
    ParseMRL( p_this);
    p_sys->open = 0;
    vlc_mutex_init( p_this, &p_sys->lock );
    vlc_cond_init( p_this, &p_sys->wait );
    msg_Dbg( p_this, "CommonOpen2...");
    if( BuildGraph( p_this, p_sys ) != 0 ) {
        msg_Err( p_this, "can't open video");
        return VLC_EGENERIC;
    } 
    msg_Info( p_this, "comp %d streams %d, norm %d fr: %d rate %d\n", p_sys->i_compression, p_sys->i_streams, p_sys->i_norm, 
        p_sys->i_framerate, p_sys->i_vrate);
    return VLC_SUCCESS;
}
 


/*****************************************************************************
 * CommonClose: close device
 *****************************************************************************/
static void CommonClose( vlc_object_t *p_this, access_sys_t *p_sys )
{
    // break all the directshow connections
    demux_t      *p_demux = (demux_t *)p_this;

    msg_Dbg( p_this, "releasing DirectShow delete");
    if( p_sys->p_capture_filter ) {
        // stop the filter
        IPin *pPin;
        p_sys->p_capture_filter->Stop();
        msg_Dbg( p_this, "releasing p_capture file\n");
        // disconnect the pins(since we don't use p_sys->p_graph
        // we can do this because the close is initiated by the Stop() of the bridge 
        // filter.
        p_sys->p_capture_filter->CustomGetAudioPin()->Disconnect();
        p_sys->p_capture_filter->CustomGetVideoPin()->Disconnect();
        pPin = (IPin *) p_demux->p_vlc->pinVid;
        pPin->Disconnect();
        pPin = (IPin *) p_demux->p_vlc->pinAud;
        pPin->Disconnect();
        p_sys->p_capture_filter->Release();
        p_sys->p_capture_filter = NULL;
    }

    if( p_sys->p_graph ) {
        msg_Dbg(p_this, "buildgraph19\n");
        msg_Info( p_this, "releasing p_graph\n");
        p_sys->p_graph->Release();
        p_sys->p_graph = NULL;
    }
    msg_Dbg( p_this, "after DirectShow delete");
    vlc_mutex_destroy( &p_sys->lock );
    vlc_cond_destroy( &p_sys->wait );
    free( p_sys );
    msg_Info( p_this, "Close");
}



/*****************************************************************************
 * DemuxClose: close device
 *****************************************************************************/
void DemuxClose( vlc_object_t *p_this )
{
    demux_t      *p_demux = (demux_t *)p_this;
    access_sys_t *p_sys   = (access_sys_t *)p_demux->p_sys;
    CommonClose( p_this, p_sys );
    p_demux->p_vlc->sraybridge_state = 0;

}


static int BuildGraph( vlc_object_t *p_this, access_sys_t *p_sys )
{
    int hr = 0;
    demux_t      *p_demux = (demux_t *)p_this;
    msg_Dbg(p_this, "buildgraph15 %x\n",hr);
    
    // we create the filter and immediately add it to the graph.
    // therefor new CaptureFilter should not addRef
    CaptureFilter *p_capture_filter = new CaptureFilter( p_this, p_sys ); // mt, media_count );
    msg_Dbg(p_this, "buildgraph16\n");
    p_sys->p_capture_filter = p_capture_filter;
    msg_Dbg(p_this, "buildgraph17\n");
    // we can't call IFilter::AddFilter BECAUSE this may be called while the FilterLock is held.
    // We have to get this up and running before the downstream graph can go out of PAUSE or
    // one will risk an 80040209 error in the Source Graph Deliver.
    p_capture_filter->JoinFilterGraph( p_sys->p_graph, 0);
    //p_sys->p_graph->AddFilter(p_capture_filter,0);
    msg_Dbg(p_this, "buildgraph18\n");
    // connect the pins 
    if( p_sys->i_streams & STREAM_MASK_VIDEO) {
        IPin *pOut;
        CapturePin *p_video_pin;
        msg_Dbg(p_this, "buildgraph20\n");
        p_video_pin = p_capture_filter->CustomGetVideoPin();
        msg_Dbg( p_this, "get video pin psys\n");
        pOut = (IPin *) p_demux->p_vlc->pinVid;
        msg_Dbg( p_this, "VIDEO base pin %p \n", pOut);
        if( pOut == NULL) {
            msg_Err( p_demux, "failed to get video pin!\n");
            return -1;
        }
        msg_Dbg( p_this, "Add ref base pin 2: %p %p\n", pOut,p_video_pin);
        // we connect the pins directly WITHOUT p_graph. This is very unconventional, but we are careful
        // and make sure there are no consequences of this.
        hr = pOut->Connect(p_video_pin, NULL);
        //hr = p_sys->p_graph->ConnectDirect(pOut, p_video_pin, NULL);
        msg_Dbg( p_this, "done connection: %x %p \n", hr, pOut);
        if( FAILED(hr)) {
            msg_Err( p_demux, "failed to connect video pin: %x!\n", hr);
          //  pOut->Release();
            return hr;
        }
        // fetch the connection media type
        AM_MEDIA_TYPE mt;
        memset(&mt,0,sizeof(AM_MEDIA_TYPE));
        // get connection media type from the graph pin
        if( p_video_pin->ConnectionMediaType( &mt) == 0) {
            // save AM_MEDIA_TYPE in easy to access form
            // without having to dereference a format block later, etc...
            if( mt.majortype == MEDIATYPE_Video) {
                if( mt.subtype == MEDIASUBTYPE_MPEG2_VIDEO) {
                    msg_Info(p_demux, "MPEG2 \n");
                    p_sys->i_compression = MPEG2;
                    msg_Info(p_demux, "MPEG2B \n");
                    if( mt.formattype == FORMAT_MPEG2Video) {
                        MPEG2VIDEOINFO *fmt = (MPEG2VIDEOINFO *) mt.pbFormat;
                        if( fmt) {
                            p_sys->i_width = fmt->hdr.bmiHeader.biWidth;
                            p_sys->i_height = fmt->hdr.bmiHeader.biHeight;
                            p_sys->i_vrate = fmt->hdr.dwBitRate;
                            msg_Dbg(p_demux, "avgtime %lld\n", fmt->hdr.AvgTimePerFrame);
                            p_sys->i_avgtimeperframe = fmt->hdr.AvgTimePerFrame;
                            p_sys->i_framerate = (int) ( 10010000000ULL/ fmt->hdr.AvgTimePerFrame);
                        } else {
                            msg_Err( p_demux, "NO FORMAT!?!\n");
                        }
                    }
                    msg_Info(p_demux, "MPEG2C \n");

                } else if (mt.subtype == MEDIASUBTYPE_MPEG1Payload) {
                    msg_Info(p_demux, "MPEG1\n");
                    p_sys->i_compression = MPEG1;
                    if( mt.formattype == FORMAT_MPEGVideo) {
                        MPEG1VIDEOINFO* fmt = (MPEG1VIDEOINFO*)mt.pbFormat;
                        if( fmt) {
                            p_sys->i_width = fmt->hdr.bmiHeader.biWidth;
                            p_sys->i_height = fmt->hdr.bmiHeader.biHeight;
                            p_sys->i_vrate = fmt->hdr.dwBitRate;
                            p_sys->i_avgtimeperframe = fmt->hdr.AvgTimePerFrame;
                            msg_Dbg(p_demux, "avgtime %lld\n", fmt->hdr.AvgTimePerFrame);
                            p_sys->i_framerate = (int) ( 10010000000ULL/ fmt->hdr.AvgTimePerFrame);
                            msg_Dbg(p_demux,"fr %d\n", p_sys->i_framerate);
                        } else {
                            msg_Err( p_demux, "NO FORMAT!?!\n");
                        }

                    }
                } else {
                    p_sys->i_compression = MPEG4; // OR Motion Jpeg
                    msg_Warn( p_demux,"MPEG4 or MJPEG VLCMux not debugged or supported yet!\n");
                    if( mt.formattype == FORMAT_VideoInfo) {
                        VIDEOINFOHEADER *pvh = (VIDEOINFOHEADER *) mt.pbFormat;
                        msg_Dbg(p_demux,"height is %ld\n", pvh->bmiHeader.biHeight);
                        msg_Dbg(p_demux,"width is %ld\n", pvh->bmiHeader.biWidth);
                        msg_Dbg(p_demux,"bitrate is %ld\n", pvh->dwBitRate);
                        p_sys->i_height = pvh->bmiHeader.biHeight;
                        p_sys->i_width = pvh->bmiHeader.biWidth;
                        p_sys->i_vrate = pvh->dwBitRate;
                        p_sys->i_framerate = (int) ( 10010000000ULL/ pvh->AvgTimePerFrame);
                        if( pvh->bmiHeader.biCompression == VLC_FOURCC( 'M', 'J', 'P', 'G')) {
                            p_sys->i_compression = MOTIONJPEG;
                        } else if( pvh->bmiHeader.biCompression == VLC_FOURCC( 'M', 'P', '4', 'S')) {
                            p_sys->i_compression = MPEG4;  // mpeg4 microsoft
                        } else if( pvh->bmiHeader.biCompression == VLC_FOURCC( 'D', 'X', '5', '0')) {
                            p_sys->i_compression = MPEG4DIVX;  // mpeg4 divx
                        } else if( pvh->bmiHeader.biCompression == VLC_FOURCC( 'D', 'I', 'V', 'X')) {
                            p_sys->i_compression = MPEG4DIVX;  // mpeg4 divx
                        } else if( pvh->bmiHeader.biCompression == VLC_FOURCC( 'd', 'i', 'v', 'x')) {
                            p_sys->i_compression = MPEG4DIVX;  // mpeg4 divx
                        }
                    } else if( mt.formattype == FORMAT_VideoInfo2) {
                        VIDEOINFOHEADER2 *pvh = (VIDEOINFOHEADER2 *) mt.pbFormat;
                        msg_Dbg(p_demux,"height is %ld\n", pvh->bmiHeader.biHeight);
                        msg_Dbg(p_demux,"width is %ld\n", pvh->bmiHeader.biWidth);
                        msg_Dbg(p_demux,"bitrate is %ld\n", pvh->dwBitRate);
                        p_sys->i_height = pvh->bmiHeader.biHeight;
                        p_sys->i_width = pvh->bmiHeader.biWidth;
                        p_sys->i_vrate = pvh->dwBitRate; 
                        p_sys->i_framerate = (int) (10010000000ULL/ pvh->AvgTimePerFrame);
                        if( pvh->bmiHeader.biCompression == VLC_FOURCC( 'M', 'J', 'P', 'G')) {
                            p_sys->i_compression = MOTIONJPEG;
                        } else if( pvh->bmiHeader.biCompression == VLC_FOURCC( 'M', 'P', '4', 'S')) {
                            p_sys->i_compression = MPEG4;  // mpeg4 microsoft
                        } else if( pvh->bmiHeader.biCompression == VLC_FOURCC( 'D', 'X', '5', '0')) {
                            p_sys->i_compression = MPEG4DIVX;  // mpeg4 divx
                        } else if( pvh->bmiHeader.biCompression == VLC_FOURCC( 'D', 'I', 'V', 'X')) {
                            p_sys->i_compression = MPEG4DIVX;  // mpeg4 divx
                        } else if( pvh->bmiHeader.biCompression == VLC_FOURCC( 'd', 'i', 'v', 'x')) {
                            p_sys->i_compression = MPEG4DIVX;  // mpeg4 divx
                        }
                    }
                }
            } else {
                msg_Err( p_demux, "Invalid connection!\n");
            }
            // ConnectionMediaType in filter.cpp AddRefs.
            msg_Dbg(p_demux,"releasing?\n");
            msg_Dbg(p_demux,"releasing done\n");
            FreeMediaType(mt);
        }
        msg_Dbg(p_demux,"Video: fr %d vrate %d\n", p_sys->i_framerate, p_sys->i_vrate);
    }

    if( p_sys->i_streams & STREAM_MASK_AUDIO) {
        IPin *pOut;
        CapturePin *p_audio_pin;
        msg_Dbg(p_this, "buildgraph19\n");
        p_audio_pin = p_capture_filter->CustomGetAudioPin();
        msg_Dbg( p_this, "get audio pin psys %p\n", p_audio_pin);
        pOut = (IPin *) p_demux->p_vlc->pinAud;
        if( pOut == NULL) {
            msg_Err( p_demux, "failed to get audio pin!\n");
            return -1;
        }
        //pOut->AddRef();
        msg_Dbg( p_this, "base pin %p \n", pOut);
        hr = pOut->Connect(p_audio_pin, NULL);
        //hr = p_sys->p_graph->ConnectDirect(pOut, p_audio_pin,NULL);
        if( FAILED(hr)) {
            msg_Err( p_this, "failed to connect audio pin %x",hr);
          //  pOut->Release();
            return hr;
        }
        // grab the media type and save the important parameters
        AM_MEDIA_TYPE mt;
        memset(&mt,0,sizeof(AM_MEDIA_TYPE));
        if( p_audio_pin->ConnectionMediaType( &mt) == 0) {
            if( mt.majortype != MEDIATYPE_Audio && mt.majortype != MEDIATYPE_Stream) {
                msg_Err(p_demux, "Invalid connection type!\n");
            }
            if( (mt.subtype != MEDIASUBTYPE_PCM) &&
                (mt.subtype != MEDIASUBTYPE_MPEG1AudioPayload) && (mt.subtype != MEDIASUBTYPE_MPEG1Audio)) {
                msg_Err(p_demux, "Invalid connection type. only uncompressed audio!\n");
            }
            if( mt.subtype == MEDIASUBTYPE_PCM) {
                p_sys->i_aud_fourcc = FOURCC_AUDIO_PCM;
            } else {
                // only other type is mpeg1 audio
                msg_Dbg(p_demux, "using mpeg1 audio\n");
                p_sys->i_aud_fourcc = FOURCC_AUDIO_MPEG1;
            }
            if( mt.formattype == FORMAT_WaveFormatEx) {
                WAVEFORMATEX *pwave = (WAVEFORMATEX *) mt.pbFormat; 
                p_sys->i_aud_channels = pwave->nChannels;
                p_sys->i_aud_sr = pwave->nSamplesPerSec;
                p_sys->i_aud_bits = pwave->wBitsPerSample;
                p_sys->i_aud_bps = pwave->nAvgBytesPerSec;
            }
            msg_Info(p_demux, "c %d sr %d %d \n", p_sys->i_aud_channels, p_sys->i_aud_sr, p_sys->i_aud_bits);
            FreeMediaType(mt);
        }
        //pOut->Release();
    }
    msg_Dbg(p_this, "buildgraph23\n");
    return 0;
}


#define MAX_MEDIA_TYPES 32


/****************************************************************************
 * Demux:
 ****************************************************************************/
static int Demux( demux_t *p_demux )
{
    access_sys_t *p_sys = (access_sys_t *)p_demux->p_sys;
    VLCMediaSample sample;
    int i_data_size, i_stream;
    static int i_frames = 0;
    static int a_frames = 0;
    static int last_vid;
    static int last_aud;
    int pts_inc=0;
    uint8_t *p_data;
    block_t *p_block;
    //msg_Info(p_demux, "Demux1");
    if( p_sys == NULL) {
        return 0;
    }
    if( !p_sys->open) {
        msg_Info(p_demux, "Demux: not open");
        return 0;
    }

    if( p_sys->p_capture_filter == NULL) {
        msg_Info(p_demux, "null cap filter demux\n");

        return 0;
    }
    //msg_Info(p_demux, "Demux3");
    vlc_mutex_lock( &p_sys->lock );
    // grab audio from the capture filter
    // unlike the dshow module, audio and video are on the same filter
    // GrabAudio
    i_stream = -1;
    if( p_sys->i_streams & STREAM_MASK_AUDIO) {
        if( p_sys->p_capture_filter->CustomGetAudioPin()->CustomGetSample( &sample) == S_OK) {
            i_stream = 0;
            a_frames--;
        }
    }

    if( i_stream == -1) {
        if( p_sys->i_streams & STREAM_MASK_VIDEO) {
            // grab video
            if( p_sys->p_capture_filter->CustomGetVideoPin()->CustomGetSample( &sample) == S_OK) {
                i_stream = 1;
                i_frames--;
            }
        }
    }
    vlc_mutex_unlock( &p_sys->lock );

    if( i_stream == -1 )  {
        /* Sleep so we do not consume all the cpu */
        /* video is at 30fps max */
        /* raw audio is coming in at 8192 byte chunks.  or 42666 */
        // 125 fps
        msleep( 8000 );
        return 1;
    }
    // got sample
    i_data_size = sample.p_sample->GetActualDataLength();
    sample.p_sample->GetPointer( &p_data );
    //msg_Info(p_demux, "demux data received: %x %x %x %x\n", p_data[0], p_data[1], p_data[2], p_data[3]);
    REFERENCE_TIME i_pts, i_end_date;
    HRESULT hr = sample.p_sample->GetTime( &i_pts, &i_end_date );

    if( hr != VFW_S_NO_STOP_TIME && hr != S_OK ) {
        i_pts = 0;
    }

    p_block = block_New( p_demux, i_data_size );
    p_demux->p_vlc->pf_memcpy( p_block->p_buffer, p_data, i_data_size );
    sample.p_sample->Release();
    if( i_stream == 0) { 
        /* audio */ 
        pts_inc = (unsigned long) ( (1000.0 * ((i_data_size * 1000.0 / p_sys->i_aud_sr) / p_sys->i_aud_blockalign)) + 0.5);
//        pts_inc = ((i_data_size * 1000) / p_sys->i_aud_bps ) * 1000;
        p_sys->m_aud_samples++;
        if( ( (i_data_size % 3) == 2 ) && (p_sys->m_aud_samples %3 == 0)) {
            /* every third sample decrement the pts_inc to round off accumulation errors. */
            /* this assumes constant data size for audio, which is always the case for */
            /* our DSHOW device(and most others) */
            pts_inc--;
        }
        last_aud = p_sys->m_aud_pts;
        p_sys->m_aud_pts += pts_inc;
        i_pts = p_sys->m_aud_pts;
    } else if (i_stream == 1) {
        pts_inc = p_sys->i_avgtimeperframe; /* times 10 DSHOW */
        last_vid = p_sys->m_vid_pts_times_ten / 10;
        p_sys->m_vid_pts_times_ten += pts_inc;
        p_sys->m_vid_samples++;
        i_pts = p_sys->m_vid_pts_times_ten / 10;
    }

#ifdef DEBUG_VERBOSE
    if( i_stream == 1) {
        msg_Dbg(p_demux, "video pts diff %ld\n", (unsigned long) (i_pts - last_vid));
    }
    if( i_stream == 0) {
        msg_Dbg(p_demux, "audio pts diff %ld size %d sr %d block %d BPS %d\n", (unsigned long) pts_inc, 
            i_data_size, p_sys->i_aud_sr,p_sys->i_aud_blockalign, p_sys->i_aud_bps);
    }
#endif

    
    p_block->i_dts = i_pts;
    p_block->i_pts = i_pts;

    es_out_Control( p_demux->out, ES_OUT_SET_PCR, (i_pts > 0) ? i_pts : 0);

    if( i_stream == 0) { 
        /* audio */ 
        es_out_Send( p_demux->out, p_sys->p_es_audio, p_block );
    } else if (i_stream == 1) {
        /* video */
        es_out_Send( p_demux->out, p_sys->p_es_video, p_block );
    }

    return 1;

}


/****************************************************************************
 * DemuxControl:
 ****************************************************************************/
static int DemuxControl( demux_t *p_demux, int i_query, va_list args )
{
    vlc_bool_t *pb;
    int64_t    *pi64;
    access_sys_t *p_sys = (access_sys_t *)p_demux->p_sys;

    switch( i_query )
    { 
    /* Special for access_demux */
    case DEMUX_CAN_PAUSE:
    case DEMUX_SET_PAUSE_STATE:
    case DEMUX_CAN_CONTROL_PACE:
        pb = (vlc_bool_t*)va_arg( args, vlc_bool_t * ); 
        *pb = VLC_FALSE;
        return VLC_SUCCESS;
    case DEMUX_GET_PTS_DELAY:
        pi64 = (int64_t*)va_arg( args, int64_t * );
        *pi64 = (int64_t)var_GetInteger( p_demux, "dshowbridge-caching" ) * 1000;
        return VLC_SUCCESS;
    case DEMUX_GET_TIME:
        pi64 = (int64_t*)va_arg( args, int64_t * ); 
        *pi64 = (int64_t) 0;
        return VLC_SUCCESS;
    /* TODO implement others */
    default:
        return VLC_EGENERIC;
    }

    return VLC_EGENERIC;
}
